from json import dumps
from os import getenv
from pkg_resources import get_distribution

from getgist import GetGistCommons
from getgist.request import GetGistRequests


def oauth_only(function):
    """Decorator to restrict some GitHubTools methods to run only with OAuth"""
    def check_for_oauth(self, *args, **kwargs):
        """
        Returns False if GitHubTools instance is not authenticated, or return
        the decorated fucntion if it is.
        """
        if not self.is_authenticated:
            self.oops('To use putgist you have to set your GETGIST_TOKEN')
            self.oops('(see `putgist --help` for details)')
            return False
        return function(self, *args, **kwargs)
    return check_for_oauth


class GitHubTools(GetGistCommons):
    """Helpers to deal with GitHub API and manipulate gists"""

    version = get_distribution('getgist').version
    api_root_url = 'https://api.github.com/'
    headers = {'Accept': 'application/vnd.github.v3+json',
               'User-Agent': 'GetGist v' + version}
    requests = GetGistRequests(headers)
    is_authenticated = False

    def __init__(self, user, filename, assume_yes=False):
        """
        Save basic variables to all methods, instantiate GetGistrequests and
        calls the OAuth method.
        :param user: (str) GitHub username
        :param filename: (str) filename to be saved (locally), created or
        updated (remotelly)
        :param assume_yes: (bool) assume yes (or first option) for all prompts
        :return: (None)
        """
        self.user = user
        self.filename = filename
        self.assume_yes = assume_yes
        self.add_oauth_header()

    def add_oauth_header(self):
        """
        Validate token and add the proper header for further requests.
        :return: (None)
        """
        # abort if no token
        oauth_token = self._get_token()
        if not oauth_token:
            return

        # add oauth header & reach the api
        self.headers['Authorization'] = 'token ' + oauth_token
        url = self._api_url('user')
        raw_resp = self.requests.get(url)
        resp = raw_resp.json()

        # abort & remove header if token is invalid
        if resp.get('login', None) != self.user:
            self.oops('Invalid token for user ' + self.user)
            self.headers.pop('Authorization')
            return

        self.is_authenticated = True
        self.yeah('User {} authenticated'.format(self.user))

    def get_gists(self):
        """
        List generator containing dictionaries gists' relevant information
        (such as id, description, filenames and raw URL).
        :return: (None)
        """
        # fetch all gists
        if self.is_authenticated:
            url = self._api_url('gists')
        else:
            url = self._api_url('users', self.user, 'gists')
        self.output('Fetching ' + url)
        raw_resp = self.requests.get(url)

        # abort if user not found
        if raw_resp.status_code != 200:
            self.oops('User `{}` not found'.format(self.user))
            return

        # abort if there are no gists
        resp = raw_resp.json()
        if not resp:
            self.oops('No gists found for user `{}`'.format(self.user))
            return

        # parse response
        for gist in raw_resp.json():
            yield self._parse_gist(gist)

    def select_gist(self, allow_none=False):
        """
        Given the requested filename, it selects the proper gist; if more than
        one gist is found with the given filename, user is asked to choose.
        :allow_none: (bool) for `getgist` it should raise error if no gist is
        found, but setting this argument to True avoid this error, which is
        useful when `putgist` is calling this method
        :return: (dict) selected gist
        """
        # pick up all macthing gists
        matches = list()
        for gist in self.get_gists():
            for gist_file in gist.get('files'):
                if self.filename == gist_file.get('filename'):
                    matches.append(gist)

        # abort if no match is found
        if not matches:
            if allow_none:
                return None
            else:
                msg = "No file named `{}` found in {}'s gists"
                self.oops(msg.format(self.filename, self.user))
                if not self.is_authenticated:
                    self.warn('To access private gists set the GETGIST_TOKEN')
                    self.warn('(see `getgist --help` for details)')
                return False

        # return if there's is only one match
        if len(matches) == 1 or self.assume_yes:
            return matches.pop(0)

        return self._ask_which_gist(matches)

    def read_gist_file(self, gist):
        """
        Returns the contents of file hosted inside a gist at GitHub.
        :param gist: (dict) gist parsed by GitHubTools._parse()
        :return: (bytes) content of a gist loaded from GitHub
        """
        url = False
        files = gist.get('files')
        for gist_file in files:
            if gist_file.get('filename') == self.filename:
                url = gist_file.get('raw_url')
                break
        if url:
            self.output('Reading {}'.format(url))
            response = self.requests.get(url)
            return response.content

    @oauth_only
    def update(self, gist, content):
        """
        Updates the contents of file hosted inside a gist at GitHub.
        :param gist: (dict) gist parsed by GitHubTools._parse_gist()
        :param content: (str or bytes) to be written
        :return: (bool) indicatind the success or failure of the update
        """
        # abort if content is False
        if content is False:
            return False

        # request
        url = self._api_url('gists', gist.get('id'))
        data = {'files': {self.filename: {'content': content}}}
        self.output('Sending contents of {} to {}'.format(self.filename, url))
        response = self.requests.patch(url, data=dumps(data))

        # error
        if response.status_code != 200:
            self.oops('Could not update ' + gist.get('description'))
            self.oops('PATCH request returned ' + str(response.status_code))
            return False

        # success
        self.yeah('Done!')
        self.hey('The URL to this Gist is: {}'.format(gist['url']))
        return True

    @oauth_only
    def create(self, content, **kwargs):
        """
        Create a new gist.
        :param gist: (dict) gist parsed by GitHubTools._parse()
        :param content: (str or bytes) to be written
        :param public: (bool) defines if the gist is public or private
        :return: (bool) indicatind the success or failure of the creation
        """
        # abort if content is False
        if content is False:
            return False

        # set new gist
        public = bool(kwargs.get('public', True))
        data = {'description': self.filename, 'public': public,
                'files': {self.filename: {'content': content}}}

        # send request
        url = self._api_url('gists')
        self.output('Sending contents of {} to {}'.format(self.filename, url))
        response = self.requests.post(url, data=dumps(data))

        # error
        if response.status_code != 201:
            self.oops('Could not create ' + self.filename)
            self.oops('POST request returned ' + str(response.status_code))
            return False

        # parse created gist
        gist = self._parse_gist(response.json())

        # success
        self.yeah('Done!')
        self.hey('The URL to this Gist is: {}'.format(gist['url']))
        return True

    def _ask_which_gist(self, matches):
        """
        Asks user which gist to use in case of more than one gist matching the
        instance filename.
        :param matches: (list) of dictioaries generated within select_gists()
        :return: (dict) of the selected gist
        """
        # ask user which gist to use
        self.hey('Use {} from which gist?'.format(self.filename))
        for count, gist in enumerate(matches, 1):
            self.hey('[{}] {}'.format(count, gist.get('description')))

        # get the gist index
        selected = False
        while not selected:
            try:
                gist_index = int(self.ask('Type the number: ')) - 1
                selected = matches[gist_index]
            except (ValueError, IndexError):
                self.oops('Invalid number, please try again.')

        self.output('Using `{}` Gist'.format(selected['description']))
        return selected

    def _api_url(self, *args):
        """Get entrypoints adding arguments separated by slashes"""
        return self.api_root_url + '/'.join(args)

    @staticmethod
    def _parse_gist(gist):
        """Receive a gist (dict) and parse it to GetGist"""

        # parse files
        files = list()
        file_names = sorted(filename for filename in gist['files'].keys())
        for name in file_names:
            files.append(dict(filename=name,
                              raw_url=gist['files'][name].get('raw_url')))

        # parse description
        description = gist['description']
        if not description:
            names = sorted(f.get('filename') for f in files)
            description = names.pop(0)

        return dict(description=description,
                    id=gist.get('id'),
                    files=files,
                    url=gist.get('html_url'))

    @staticmethod
    def _get_token():
        """Retrieve username from env var"""
        return getenv('GETGIST_TOKEN')



import requests


class GetGistRequests(object):
    """Encapsulate requests lib to always send self.headers as headers"""

    def __init__(self, headers=None):
        """
        Get a header object to use it in all requests
        :param headers: (dict)
        :return: (None)
        """
        if not headers:
            headers = dict()
        self.headers = headers

    def add_headers(self, kwargs):
        """
        Add any extra header to the existng header object.
        :param kwargs: (dict)
        :return: (dict) containing an item with `headers` as key
        """
        kwargs['headers'] = kwargs.get('headers', {})
        kwargs['headers'].update(self.headers)
        return kwargs

    def get(self, url, params=None, **kwargs):
        """Encapsulte requests.get to use this class instance header"""
        return requests.get(url, params=params, **self.add_headers(kwargs))

    def patch(self, url, data=None, **kwargs):
        """Encapsulte requests.patch to use this class instance header"""
        return requests.patch(url, data=data, **self.add_headers(kwargs))

    def post(self, url, data=None, **kwargs):
        """Encapsulte requests.post to use this class instance header"""
        return requests.post(url, data=data, **self.add_headers(kwargs))



from sys import stderr, stdout, version_info

from click import secho

input_method = input
if version_info < (3, 0):
    input_method = raw_input


class GetGistCommons(object):
    """Basic output methods used to print messages on users' terminal"""

    indent_size = 2
    indent_char = ' '

    def indent(self, message):
        """
        Sets the indent for standardized output
        :param message: (str)
        :return: (str)
        """
        indent = self.indent_char * self.indent_size
        return indent + message

    def output(self, message, color=None):
        """
        A helper to used like print() or click's secho() tunneling all the
        outputs to sys.stdout or sys.stderr
        :param message: (str)
        :param color: (str) check click.secho() documentation
        :return: (None) prints to sys.stdout or sys.stderr
        """
        output_to = stderr if color == 'red' else stdout
        secho(self.indent(message), fg=color, file=output_to)

    def ask(self, message):
        """
        A helper to indent input()
        :param message: (str)
        :return: (str) the user input
        """
        return input_method(self.indent(message))

    def oops(self, message):
        """Helper to colorize error messages"""
        return self.output(message, color='red')

    def yeah(self, message):
        """Helper to colorize success messages"""
        return self.output(message, color='green')

    def warn(self, message):
        """Helper to colorize warning messages"""
        return self.output(message, color='yellow')

    def hey(self, message):
        """Helper to colorize highlihghted messages"""
        return self.output(message, color='blue')



from os import getenv

from click import argument, command, option

from getgist.github import GitHubTools
from getgist.local import LocalTools


GETGIST_DESC = """
    GetGist downloads any file from a GitHub Gist, with one single command.
    Usage: `getgist <GitHub username> <file name from any file inside a gist>`.

    If you set GETGIST_USER envvar with your GitHub username, you can use the
    shortcut `geymy <file name>` (see `getmy --help` for details).

    If you set GETGIST_TOKEN envvar with your personal access token (see
    https://github.com/settings/tokens for details) you can get get priavte
    gists from your account and you can upload local changes to your gist repo
    (see `putmy --help` for details).
"""


GETMY_DESC = """
    Call `getgist` assuming the user is set in an envvar called GETGIST_USER.
    See `getgist --help` for more details.
"""

PUTGIST_DESC = """
    PutGist uploads any file to a GitHub Gist, with one single command.
    Usage: `putgist <GitHub username> <file name>`.

    You have to set the GETGIST_TOKEN envvar with your personal access token
    (see https://github.com/settings/tokens for details).

    If you set GETGIST_USER envvar with your GitHub username, you can use the
    shortcut `putmy <file name>` (see `getmy --help` for details).
"""


PUTMY_DESC = """
    Call `putgist` assuming the user is set in an envvar called GETGIST_USER.
    See `putgist --help` for more details.
"""


class GetGist(object):
    """
    Main GetGist objects linking inputs from the CLI to the helpers from
    GitHubTools (to deal with the API) and LocalTools (to deal with the local
    file system.
    """

    def __init__(self, **kwargs):
        """
        Instantiate GitHubTools & LocalTools, and set the variables required
        to get, create or update gists (filename and public/private flag)
        :param user: (str) GitHub username
        :param filename: (str) name of file from any Gist or local file system
        :param allow_none: (bool) flag to use GitHubTools.select_gist
        differently with `getgist` and `putgist` commands (if no gist/filename
        is found it raises an error for `getgist`, or sets `putgist` to create
        a new gist).
        :param create_private: (bool) create a new gist as private
        :param assume_yes: (bool) assume yes (or first option) for all prompts
        :return: (None)
        """

        # get arguments
        user = kwargs.get('user')
        allow_none = kwargs.get('allow_none', False)
        assume_yes = kwargs.get('assume_yes', False)
        filename = kwargs.get('filename')
        self.public = not kwargs.get('create_private', False)

        # instantiate local tools & check for user
        self.local = LocalTools(filename, assume_yes)
        if not user:
            message = """
                No default user set yet. To avoid this prompt set an
                environmental variable called  `GETGIST_USER`.'
            """
            self.local.oops(message)

        # instantiate filename, guthub tools and fetch gist
        self.github = GitHubTools(user, filename, assume_yes)
        self.gist = self.github.select_gist(allow_none)

    def get(self):
        """Reads the remote file from Gist and save it locally"""
        if self.gist:
            content = self.github.read_gist_file(self.gist)
            self.local.save(content)

    def put(self):
        """ Reads local file & update the remote gist (or create a new one)"""
        content = self.local.read(self.github.filename)
        if self.gist:
            self.github.update(self.gist, content)
        else:
            self.github.create(content, public=self.public)


@command(help=GETGIST_DESC)
@option('--yes-to-all', '-y', is_flag=True, help='Assume yes to all prompts.')
@argument('user')
@argument('filename')
def run_getgist(filename, user, **kwargs):
    """Passes user inputs to GetGist() and calls get()"""
    assume_yes = kwargs.get('yes_to_all')
    getgist = GetGist(user=user, filename=filename, assume_yes=assume_yes)
    getgist.get()


@command(help=GETMY_DESC)
@option('--yes-to-all', '-y', is_flag=True, help='Assume yes to all prompts.')
@argument('filename')
def run_getmy(filename, **kwargs):
    """Shortcut for run_getgist() reading username from env var"""
    assume_yes = kwargs.get('yes_to_all')
    user = getenv('GETGIST_USER')
    getgist = GetGist(user=user, filename=filename, assume_yes=assume_yes)
    getgist.get()


@command(help=PUTGIST_DESC)
@option('--yes-to-all', '-y', is_flag=True, help='Assume yes to all prompts.')
@option('--private', '-p', is_flag=True, help='Crete new gist as private')
@argument('user')
@argument('filename')
def run_putgist(filename, user, **kwargs):
    """Passes user inputs to GetGist() and calls put()"""
    assume_yes = kwargs.get('yes_to_all')
    private = kwargs.get('private')
    getgist = GetGist(user=user, filename=filename, assume_yes=assume_yes,
                      create_private=private, allow_none=True)
    getgist.put()


@command(help=PUTMY_DESC)
@option('--yes-to-all', '-y', is_flag=True, help='Assume yes to all prompts.')
@option('--private', '-p', is_flag=True, help='Crete new gist as private')
@argument('filename')
def run_putmy(filename, **kwargs):
    """Shortcut for run_putgist() reading username from env var"""
    assume_yes = kwargs.get('yes_to_all')
    private = kwargs.get('private')
    user = getenv('GETGIST_USER')
    getgist = GetGist(user=user, filename=filename, assume_yes=assume_yes,
                      create_private=private, allow_none=True)
    getgist.put()



import os

from getgist import GetGistCommons


class LocalTools(GetGistCommons):
    """Helpers to deal with local files and local file system"""

    def __init__(self, filename, assume_yes=False):
        """
        Sets the file name to be used by the instance.
        :param filename: (str) local file name (ro be read or written)
        :param assume_yes: (bool) assume yes (or first option) for all prompts
        return: (None)
        """
        self.cwd = os.getcwd()
        self.filename = filename
        self.assume_yes = assume_yes
        self.path = os.path.join(self.cwd, filename)

    def save(self, content):
        """
        Save any given content to the instance file.
        :param content: (str or bytes)
        :return: (None)
        """
        # backup existing file if needed
        if os.path.exists(self.path) and not self.assume_yes:
            message = 'Overwrite existing {}? (y/n) '
            confirm = self.ask(message.format(self.filename))
            if confirm.lower() != 'y':
                self.backup()

        # write file
        self.output('Saving ' + self.filename)
        with open(self.path, 'wb') as handler:
            if not isinstance(content, bytes):
                content = bytes(content, 'utf-8')
            handler.write(content)
        self.yeah('Done!')

    def backup(self):
        """Backups files with the same name of the instance filename"""
        count = 0
        name = '{}.bkp'.format(self.filename)
        backup = os.path.join(self.cwd, name)
        while os.path.exists(backup):
            count += 1
            name = '{}.bkp{}'.format(self.filename, count)
            backup = os.path.join(self.cwd, name)
        self.hey('Moving existing {} to {}'.format(self.filename, name))
        os.rename(os.path.join(self.cwd, self.filename), backup)

    def read(self, filename=None):
        """
        Read the contents of a file.
        :param filename: (str) path to a file in the local file system
        :return: (str) contents of the file, or (False) if not found/not file
        """
        if not filename:
            filename = self.path

        # abort if the file path does not exist
        if not os.path.exists(filename):
            self.oops('Sorry, but {} does not exist locally'.format(filename))
            return False

        # abort if the file path is not a file
        if not os.path.isfile(filename):
            self.oops('Sorry, but {} is not a file'.format(filename))
            return False

        with open(filename) as handler:
            return handler.read()



from setuptools import setup

REPO_URL = 'http://github.com/cuducos/getgist'


setup(author='Eduardo Cuducos',
      author_email='cuducos@gmail.com',
      classifiers=['Development Status :: 4 - Beta',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Intended Audience :: Developers',
                   'Topic :: Utilities',
                   'License :: OSI Approved :: MIT License'],
      description='CLI to update local and remote files from GitHub gists',
      entry_points={'console_scripts': ['getgist=getgist.__main__:run_getgist',
                                        'getmy=getgist.__main__:run_getmy',
                                        'putgist=getgist.__main__:run_putgist',
                                        'putmy=getgist.__main__:run_putmy']},
      include_package_data=True,
      install_requires=['click>=5.1', 'requests>=2.10.0'],
      keywords='gist, command-line, github, dotfiles',
      license='MIT',
      long_description='Check `GetGist at GitHub <{}>`_.'.format(REPO_URL),
      name='getgist',
      packages=['getgist'],
      setup_requires=['tox-setuptools'],
      url=REPO_URL,
      version='0.1.1',
      zip_safe=False)
